
/* A5SES is released under the zlib license.
 * See LICENSE.txt for more info.
 */

/// @file Atlas.hpp
/// @ingroup Atlas
/// Texure atlasing

#ifndef A5SES_ATLAS_HPP_INCLUDED
#define A5SES_ATLAS_HPP_INCLUDED

#include "common.hpp"
#include "Bitmap.hpp"

#include <vector>
#include <set>

namespace a5
{

class Atlas_Bitmap;

/// @ingroup Atlas
/// Texture atlas
class Atlas
{
	private:
		int width;
		int height;
		int x_res;
		int y_res;
		std::vector<Bitmap *> pages;
		std::set<Atlas_Bitmap *> children;
		std::vector<std::vector<char> > alloc_map; /* vector<bool> is slow in C++98 */
		bool destroying;
		
		void Free(Atlas_Bitmap *ref);

		// Disable copying
		Atlas(const Atlas &);
		Atlas &operator =(const Atlas &);

	public:
		/// @ingroup Atlas Event
		/// Thrown when trying to allocate a too-large section of the atlas
		class Too_Big : public Exception
		{
			public:
				virtual const char *what()
				{
					return "a5::Event_Reader::Too_Big";
				}

				virtual ~Too_Big() throw() { }
		};

		/// Creates a texture atlas with the specified dimensions and resolution
		Atlas(int width_, int height_, int x_res_ = 16, int y_res_ = 16)
			: width(width_)
			, height(height_)
			, x_res(x_res_)
			, y_res(y_res_)
			, destroying(false)
		{
			NewPage();
		}

		/// Creates a new page on the atlas
		void NewPage();

		/// Allocates room for and copies an existing bitmap in to the atlas
		A5SES_PTR_TYPE<Bitmap> Add(Bitmap &bitmap);
		
		/// Allocates an image to the atlas with the specified width/height
		A5SES_PTR_TYPE<Bitmap> Alloc(int width, int height);

		/// Returns the number of pages in the atlas
		int Pages()
		{
			return pages.size();
		}

		/// Returns the atlas bitmap for the given page
		a5::Bitmap &GetPage(int page)
		{
			if (page < 0 || page >= Pages())
				throw Exception();

			return *pages[page];
		}
		
		/// Frees all of the allocated bitmaps in the atlas
		void Clear(bool delete_pages = false)
		{
			for (size_t page = 0; page < alloc_map.size(); ++page)
			{
				for (std::vector<char>::iterator it = alloc_map[page].begin(); it != alloc_map[page].end(); ++it)
				{
					*it = 0;
				}
				
				GetPage(page).Clear();
			}
			
			if (delete_pages)
			{
				for (size_t page = 1; page < alloc_map.size(); ++page)
				{
					delete pages[page];
				}
				
				pages.resize(1);
			}
		}
		
		/// Destroys all of the atlas pages and references
		~Atlas();
	
	friend class a5::Atlas_Bitmap;
};

/// @ingroup Atlas
/// Bitmap containing reference to an allocated part of a texture atlas
class Atlas_Bitmap : public a5::Bitmap
{
	private:
		a5::Atlas *atlas;
		int page;
		int block_x;
		int block_y;

	public:
		Atlas_Bitmap(a5::Atlas *atlas_, ALLEGRO_BITMAP *bitmap, int page_, int block_x_, int block_y_)
			: a5::Bitmap(bitmap, false)
			, atlas(atlas_)
			, page(page_)
			, block_x(block_x_)
			, block_y(block_y_)
		{ }

		/// Unregisters the bitmap from the atlas
		~Atlas_Bitmap()
		{
			atlas->Free(this);
		}
	
	friend class a5::Atlas;
};

}

#endif // A5SES_ATLAS_HPP_INCLUDED
